package JointFeldmanProtocol;

import Communication.MailHandler;
import Communication.Network;
import Communication.User;
import ShamirSecretSharing.Polynomial;
import UserInterface.DistributedKeyGenerationUser;
import com.google.protobuf.ByteString;
import com.google.protobuf.Message;
import meerkat.protobuf.DKGMessages;
import org.factcenter.qilin.primitives.Group;

import java.math.BigInteger;
import java.util.Arrays;
import java.util.Set;
import JointFeldmanProtocol.DistributedKeyGeneration.ComplainState;

/**
 * Created by Tzlil on 3/14/2016.
 */
public class DistributedKeyGenerationUserImpl implements DistributedKeyGenerationUser {

    protected final DistributedKeyGeneration dkg;

    protected final BigInteger g;
    protected final Group<BigInteger> group;
    protected final int n;
    protected final int t;
    protected final int id;

    protected MessageHandler messageHandler;
    protected final User user;
    protected final DistributedKeyGenerationParty[] parties;
    protected Set<Integer> QUAL; // set of all non-disqualified parties
    protected BigInteger[] commitments; // public verification values
    protected Polynomial.Point share; // final share of the secrete
    protected BigInteger y; // final public value

    public DistributedKeyGenerationUserImpl(DistributedKeyGeneration dkg, Network network){
        this(dkg,network,new DistributedKeyGenerationMailHandler(null));
    }
    public DistributedKeyGenerationUserImpl(DistributedKeyGeneration dkg, Network network, MailHandler mailHandler) {
        this.dkg = dkg;

        this.g = dkg.getGenerator();
        this.group = dkg.getGroup();
        this.n = dkg.getN();
        this.t = dkg.getT();
        this.id = dkg.getId();

        this.messageHandler = new MessageHandler();
        mailHandler.setMessageHandler(this.messageHandler);
        this.user = network.connect(mailHandler,dkg.getId());
        this.parties = dkg.getParties();

        this.parties[id - 1].share = dkg.getShare(id);

        this.QUAL = null;
        this.commitments = null;
        this.share = null;
        this.y = null;
    }

    /**
     * stage1 according to the protocol
     *  1. Pi broadcasts Aik for k = 0,...,t.
     *  2. Pi computes the shares Sij for j = 1,...,n and sends Sij secretly to Pj.
     */
    protected void stage1() {
        dkg.broadcastCommitments(user);
        dkg.sendSecrets(user);
    }

    /**
     * stage2 according to the protocol
     *  Pj verifies all the shares he received
     *  if check fails for an index i, Pj broadcasts a complaint against Pi.
     *  Pj broadcasts done message at the end of this stage
     */
    protected void stage2(){
        Polynomial.Point[] shares = new Polynomial.Point[n];
        BigInteger[][] commitmentsTable = new BigInteger[n][];
        for (int i = 0 ; i < n ; i++){
            shares[i] = parties[i].share;
            commitmentsTable[i] = parties[i].commitments;
        }
        dkg.broadcastComplains(user);
        //broadcast done message after all complaints
        DKGMessages.EmptyMessage doneMessage = DKGMessages.EmptyMessage.newBuilder().build();
        user.broadcast(DKGMessages.Mail.Type.DONE,doneMessage);
    }

    /**
     * stage3 according to the protocol
     *  1. if more than t players complain against a player Pi he is disqualified.
     *     otherwise Pi broadcasts the share Sij for each complaining player Pj.
     *  2. if any of the revealed shares fails the verification test, player Pi is disqualified.
     *     set QUAL to be the set of non-disqualified players.
     */
    protected void stage3(){
        dkg.answerAllComplainingPlayers(user);

        // wait until there is no complaint waiting for answer
        for (int i = 0; i < n; i++){
            for (int j = 0; j < n; j++){
                while (parties[i].complaints[j].equals(ComplainState.Waiting) && !parties[i].aborted){
                    try {
                        Thread.sleep(300);
                    } catch (InterruptedException e) {
                        // do nothing
                    }
                }
            }
        }
        this.QUAL = dkg.calcQUAL();
    }

    /**
     * stage4 according to the protocol
     *  1. public value y is computed as y = multiplication of yi mod p for i in QUAL
     *  2. public verification values are computed as Ak = multiplication of Aik mod p for i in QUAL for k = 0,...,t
     *  3. Pj sets is share of the secret as xj = sum of Sij mod q for i in QUAL
     */
    protected void stage4(){
        this.y = dkg.calcY(QUAL);
        this.commitments = dkg.calcCommitments(QUAL);
        this.share = dkg.calcShare(QUAL);
    }

    protected void endOfStage1(){
        for (int i = 0 ; i < n ; i++){
            while (parties[i].share == null && !parties[i].aborted){
                try {
                    Thread.sleep(300);
                } catch (InterruptedException e) {
                    // do nothing
                }
            }
        }

        for (int i = 0 ; i < n ; i++){
            for (int k = 0 ; k <=  t ; k++) {
                while (parties[i].commitments[k] == null && !parties[i].aborted) {
                    try {
                        Thread.sleep(300);
                    } catch (InterruptedException e) {
                        // do nothing
                    }
                }
            }
        }
    }

    @Override
    public void run() {
        user.getReceiverThread().start();
        stage1();
        endOfStage1();
        stage2();
        for (int i = 0 ; i < n ; i++){
            while (!parties[i].doneFlag && !parties[i].aborted){
                try {
                    Thread.sleep(300);
                } catch (InterruptedException e) {
                    // do nothing
                }
            }
        }

        stage3();
        stage4();
        user.getReceiverThread().interrupt();
    }

    @Override
    public BigInteger[] getCommitments() {
        return Arrays.copyOf(commitments, commitments.length);
    }

    @Override
    public BigInteger getGenerator() {
        return g;
    }

    @Override
    public Group<BigInteger> getGroup() {
        return group;
    }

    @Override
    public Polynomial.Point getShare() {
        return share;
    }

    @Override
    public int getID() {
        return id;
    }

    @Override
    public int getN() {
        return n;
    }

    @Override
    public int getT() {
        return t;
    }

    @Override
    public BigInteger getPublicValue() {
        return y;
    }

    @Override
    public Set<Integer> getQUAL() {
        return QUAL;
    }

    protected class MessageHandler implements Communication.MessageHandler{
        protected boolean isValidComplaintMessage(int sender, boolean isBroadcast, DKGMessages.IDMessage complaintMessage){
            int i = sender;
            int j = complaintMessage.getId();
            return isBroadcast && parties[i - 1].complaints[j - 1].equals( ComplainState.Non);
        }

        @Override
        public void handelComplaintMessage(int sender, boolean isBroadcast, Message message) {
            DKGMessages.IDMessage complaintMessage = (DKGMessages.IDMessage)message;
            if(isValidComplaintMessage(sender,isBroadcast,complaintMessage)){
                int i = sender;
                int j = complaintMessage.getId();
                parties[j - 1].complaints[i - 1] = ComplainState.Waiting;
            }
        }

        protected boolean isValidDoneMessage(int sender, boolean isBroadcast){
            return isBroadcast && !parties[sender - 1].doneFlag;
        }

        @Override
        public void handelDoneMessage(int sender, boolean isBroadcast,Message message) {
            if(isValidDoneMessage(sender,isBroadcast)) {
                parties[sender - 1].doneFlag = true;
            }
        }

        protected boolean isValidCommitmentMessage(int sender, boolean isBroadcast, DKGMessages.CommitmentMessage commitmentMessage){
            int i = sender - 1;
            int k = commitmentMessage.getK();
            return isBroadcast && parties[i].commitments[k] == null;
        }

        @Override
        public void handelCommitmentMessage(int sender, boolean isBroadcast, Message message) {
            DKGMessages.CommitmentMessage commitmentMessage = (DKGMessages.CommitmentMessage) message;
            if(isValidCommitmentMessage(sender,isBroadcast,commitmentMessage)){
                int i = sender - 1;
                int k = commitmentMessage.getK();
                parties[i].commitments[k] = extractCommitment(commitmentMessage);
            }
        }

        protected boolean isValidSecretMessage(int sender, boolean isBroadcast, DKGMessages.SecretMessage secretMessage){
            int i = secretMessage.getI();
            int j = secretMessage.getJ();
            if(sender != i  || isBroadcast)
                return false;
            else
                return parties[i - 1].share == null && j == id;

        }

        @Override
        public void handelSecretMessage(int sender, boolean isBroadcast, Message message) {
            DKGMessages.SecretMessage secretMessage = (DKGMessages.SecretMessage) message;
            if(isValidSecretMessage(sender,isBroadcast,secretMessage)) {
                int i = secretMessage.getI();
                Polynomial.Point secret = extractSecret(id,secretMessage.getSecret());
                parties[i - 1].share = secret;
            }
        }

        protected boolean isValidAnswerMessage(int sender, boolean isBroadcast, DKGMessages.SecretMessage secretMessage){
            int i = secretMessage.getI();
            int j = secretMessage.getJ();
            if(sender != i  || !isBroadcast)
                return false;
            else
                return j >= 1 && j <= n && parties[i - 1].complaints[j - 1].equals(ComplainState.Waiting);
        }

        @Override
        public void handelAnswerMessage(int sender, boolean isBroadcast, Message message) {
            DKGMessages.SecretMessage secretMessage = (DKGMessages.SecretMessage) message;
            if(isValidAnswerMessage(sender,isBroadcast,secretMessage)) {
                int i = secretMessage.getI();
                int j = secretMessage.getJ();
                Polynomial.Point secret = extractSecret(j,secretMessage.getSecret());
                if (dkg.isValidSecret(secret, parties[i - 1].commitments, j))
                    parties[i - 1].complaints[j - 1] = ComplainState.NonDisqualified;
                else
                    parties[i - 1].complaints[j - 1] = ComplainState.Disqualified;
            }
        }

        @Override
        public void handelAbortMessage(int sender, boolean isBroadcast, Message message) {
            parties[sender - 1].aborted = true;
        }

        public Polynomial.Point extractSecret(int i, ByteString secret){
            BigInteger x = BigInteger.valueOf(i);
            BigInteger y = new BigInteger(secret.toByteArray());
            return new Polynomial.Point(x,y);
        }
        public BigInteger extractCommitment(DKGMessages.CommitmentMessage commitmentMessage){
            return new BigInteger(commitmentMessage.getCommitment().toByteArray());
        }
    }
}
