package SecureDistributedKeyGenerationForDiscreteLogBasedCryptosystem;

import Arithmetics.Arithmetic;
import Arithmetics.Fp;
import Communication.Network;
import JointFeldmanProtocol.DistributedKeyGeneration;
import JointFeldmanProtocol.DistributedKeyGenerationUserImpl;
import ShamirSecretSharing.Polynomial;
import ShamirSecretSharing.SecretSharing;
import com.google.protobuf.Message;
import meerkat.protobuf.DKGMessages;

import java.math.BigInteger;

/**
 * Created by Tzlil on 3/16/2016.
 */
public class SecureDistributedKeyGenerationUserImpl extends DistributedKeyGenerationUserImpl {

    private final SecureDistributedKeyGeneration sdkg;
    private SecureDistributedKeyGenerationParty[] parties;
    private Arithmetic<BigInteger> arithmetic;

    public SecureDistributedKeyGenerationUserImpl(SecureDistributedKeyGeneration sdkg, Network network) {
        super(sdkg, network,new SecureDistributedKeyGenerationMailHandler(null));
        this.sdkg = sdkg;
        this.messageHandler = new MessageHandler();
        this.user.setMessageHandler(this.messageHandler);
        this.parties = sdkg.getParties();
        this.arithmetic = new Fp(sdkg.getQ());
    }

    /**
     * stage1 according to the protocol
     *  1. Pi broadcasts Cik=Aik*Bik for k = 0,...,t.
     *  2. Pi computes the shares Sij,Sij' for j = 1,...,n and sends Sij,Sij' secretly to Pj.
     */
    @Override
    protected void stage1() {
        sdkg.broadcastVerificationValues(user);
        sdkg.sendSecrets(user);
    }

    @Override
    protected void endOfStage1(){
        super.endOfStage1();
        BigInteger[] temp;
        for (int i = 0 ; i < n; i++){
            temp = parties[i].verifiableValues;
            parties[i].verifiableValues = parties[i].commitments;
            parties[i].commitments = temp;
        }
    }

    /**
     * stage2 according to the protocol
     *  Pj verifies all the shares,sharesT he received
     *  if check fails for an index i, Pj broadcasts a complaint against Pi.
     *  Pj broadcasts done message at the end of this stage
     */
    @Override
    protected void stage2(){
        sdkg.broadcastComplains(user);
        //broadcast done message after all complaints
        DKGMessages.EmptyMessage doneMessage = DKGMessages.EmptyMessage.newBuilder().build();
        isVerificationValue = false;
        user.broadcast(DKGMessages.Mail.Type.DONE,doneMessage);
    }

    private void ys(){
        sdkg.broadcastCommitments(user);
        //wait for receive all commitments from all i in QUAL
        for (int i:QUAL) {
            for(int k = 0; k <= t; k++) {
                while (parties[i - 1].commitments[k] == null && !parties[i - 1].aborted) {
                    try {
                        Thread.sleep(300);
                    } catch (InterruptedException e) {
                        // do nothing
                    }
                }
            }
        }
        sdkg.broadcastComplaints(user,true);
        //broadcast done message after all complaints
        DKGMessages.EmptyMessage doneMessage = DKGMessages.EmptyMessage.newBuilder().build();
        user.broadcast(DKGMessages.Mail.Type.DONE,doneMessage);

        for (int i:QUAL) {
            while (parties[i - 1].doneFlag && !parties[i - 1].aborted) {
                try {
                    Thread.sleep(300);
                } catch (InterruptedException e) {
                    // do nothing
                }
            }
        }

        int counter = 0;
        for (int i:QUAL) {
            if(parties[i - 1].aborted){
                counter++;
                sdkg.broadcastAnswer(user, parties[i - 1].share, parties[i - 1].shareT, i);
            }
        }
        for (int i:QUAL) {
            if(parties[i - 1].aborted){
                while (parties[i - 1].restoreSharesSet.size() < n - counter) {
                    try {
                        Thread.sleep(300);
                    } catch (InterruptedException e) {
                        // do nothing
                    }
                }
            }
        }

        for (int i = 0; i < n ; i++) {
            if(parties[i].restoreSharesSet.isEmpty()){
                continue;
            }
            try {
                Polynomial.Point[] shares = new Polynomial.Point[parties[i].restoreSharesSet.size()];
                parties[i].restoreSharesSet.toArray(shares);
                Polynomial polynomial = SecretSharing.restorePolynomial(shares,arithmetic);
                BigInteger[] coefficients = polynomial.getCoefficients();
                for (int k = 0 ; k <= t; k++){
                    parties[i].commitments[k] = group.multiply(g,coefficients[k]);
                }
                parties[i].share = new Polynomial.Point(BigInteger.valueOf(id),polynomial);

            } catch (Exception e) {
                //
            }
        }
    }

    @Override
    protected void stage4() {
        isStage4 = true;
        ((SecureDistributedKeyGenerationMailHandler)user.getMailHandler()).setStage4(true);
        ys();
        super.stage4();
    }
    boolean isStage4 = false;
    boolean isVerificationValue = true;
    private class MessageHandler extends DistributedKeyGenerationUserImpl.MessageHandler{

        protected boolean isValidSecretMessage(int sender, boolean isBroadcast, DKGMessages.DoubleSecretMessage doubleSecretMessage) {
            DKGMessages.SecretMessage secretMessage = DKGMessages.SecretMessage.newBuilder()
                    .setI(doubleSecretMessage.getI())
                    .setJ(doubleSecretMessage.getJ())
                    .setSecret(doubleSecretMessage.getSecret())
                    .build();
            return super.isValidSecretMessage(sender,isBroadcast,secretMessage);
        }

        @Override
        public void handelSecretMessage(int sender, boolean isBroadcast, Message message) {
            DKGMessages.DoubleSecretMessage doubleSecretMessage = (DKGMessages.DoubleSecretMessage)message;
            if (isValidSecretMessage(sender,isBroadcast,doubleSecretMessage)) {
                int i = doubleSecretMessage.getI();

                parties[i - 1].share = extractSecret(id, doubleSecretMessage.getSecret());
                parties[i - 1].shareT = extractSecret(id, doubleSecretMessage.getSecretT());
            }
        }
        protected boolean isValidAnswerMessage(int sender, boolean isBroadcast, DKGMessages.DoubleSecretMessage doubleSecretMessage) {
            if(!isStage4) {
                DKGMessages.SecretMessage secretMessage = DKGMessages.SecretMessage.newBuilder()
                        .setI(doubleSecretMessage.getI())
                        .setJ(doubleSecretMessage.getJ())
                        .setSecret(doubleSecretMessage.getSecret())
                        .build();
                return super.isValidAnswerMessage(sender, isBroadcast, secretMessage);
            }else{
                int i = doubleSecretMessage.getI();
                int j = doubleSecretMessage.getJ();
                return isBroadcast && j == sender && QUAL.contains(i) && QUAL.contains(j);
            }
        }

        @Override
        public void handelAnswerMessage(int sender, boolean isBroadcast, Message message) {
            DKGMessages.DoubleSecretMessage doubleSecretMessage = (DKGMessages.DoubleSecretMessage)message;
            if(isValidAnswerMessage(sender,isBroadcast,doubleSecretMessage)) {
                int i = doubleSecretMessage.getI();
                int j = doubleSecretMessage.getJ();
                Polynomial.Point secret = extractSecret(j, doubleSecretMessage.getSecret());
                Polynomial.Point secretT = extractSecret(j, doubleSecretMessage.getSecretT());
                if (!isStage4) {
                    if (sdkg.isValidSecret(secret, secretT, parties[j - 1].verifiableValues, i)) {
                        parties[i - 1].complaints[j - 1] = DistributedKeyGeneration.ComplainState.NonDisqualified;
                    } else {
                        parties[i - 1].complaints[j - 1] = DistributedKeyGeneration.ComplainState.Disqualified;
                    }
                } else {
                    parties[i - 1].restoreSharesSet.add(secret);
                }
            }
        }

        @Override
        protected boolean isValidDoneMessage(int sender, boolean isBroadcast) {
            if(!isStage4) {
                return super.isValidDoneMessage(sender, isBroadcast);
            }else{
                return isBroadcast &&  parties[sender - 1].doneFlag;
            }
        }

        @Override
        public void handelDoneMessage(int sender, boolean isBroadcast, Message message) {
            if(!isStage4)
                super.handelDoneMessage(sender, isBroadcast, message);
            else{
                if(isValidDoneMessage(sender,isBroadcast)) {
                    parties[sender - 1].doneFlag = false;
                }
            }
        }

        protected boolean isValidComplaintMessage(int sender, boolean isBroadcast,
                                                  DKGMessages.DoubleSecretMessage ysComplaintMessage){
            int i = ysComplaintMessage.getI();
            int j = ysComplaintMessage.getJ();
            return isBroadcast && j == sender && QUAL.contains(i) && QUAL.contains(j);
        }



        @Override
        public void handelComplaintMessage(int sender, boolean isBroadcast, Message message) {
            if(!isStage4) {
                super.handelComplaintMessage(sender, isBroadcast, message);
            }else {
                DKGMessages.DoubleSecretMessage ysComplaintMessage =(DKGMessages.DoubleSecretMessage)message;
                if (isValidComplaintMessage(sender,isBroadcast,ysComplaintMessage)) {
                    int i = ysComplaintMessage.getI();
                    int j = ysComplaintMessage.getJ();
                    Polynomial.Point secret = extractSecret(i,ysComplaintMessage.getSecret());
                    Polynomial.Point secretT = extractSecret(i,ysComplaintMessage.getSecretT());
                    if (sdkg.isValidSecret(secret, secretT, parties[i - 1].commitments, j)
                            && !sdkg.isValidSecret(secret,parties[i - 1].commitments, j)) {
                        parties[i - 1].restoreSharesSet.add(secret);
                        sdkg.broadcastAnswer(user, secret, secretT, i);
                    }
                }
            }
        }
    }
}
